/* $Id: types.c,v 1.12 2009/01/24 15:33:55 mit-sato Exp $ */

#include <assert.h>
#include <string.h>
#include "toy.h"
#include "interp.h"
#include "types.h"
#include "error.h"
#include "global.h"

Toy_Type*
new_nil() {
    Toy_Type *o;
    o = GC_MALLOC_ATOMIC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = NIL;
    return o;
}

Toy_Type*
new_symbol(char *symbol) {
    Toy_Type *o;
    char *p;
    int l;

    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = SYMBOL;
    o->u.symbol.cell = new_cell(symbol);
    o->u.symbol.hash_index = hash_string_key(symbol);

    p = cell_get_addr(o->u.symbol.cell);
    l = cell_get_length(o->u.symbol.cell);
    if (p[l-1] == ':') {
	o->tag |= TAG_NAMED_MASK;
    }
    if (p[0] == ':') {
	o->tag |= TAG_SWITCH_MASK;
    }

    return o;
}

Toy_Type*
new_ref(char *ref) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = REF;
    o->u.ref.cell = new_cell(ref);
    o->u.ref.hash_index = hash_string_key(ref);
    return o;
}

Toy_Type*
new_integer(long long int integer) {
    Toy_Type *o;
    o = GC_MALLOC_ATOMIC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = INTEGER;
    o->u.integer = integer;
    return o;
}

Toy_Type*
new_real(double real) {
    Toy_Type *o;
    o = GC_MALLOC_ATOMIC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = REAL;
    o->u.real = real;
    return o;
}

Toy_Type*
new_string_str(char *string) {
    Toy_Type *o;
    Cell *c;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;
    c = new_cell(string);
    if (NULL == c) return NULL;

    o->tag = STRING;
    o->u.string = c;
    return o;
}

Toy_Type*
new_string_cell(Cell *string) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = STRING;
    o->u.string = string;
    return o;
}

Toy_Type*
new_script(Toy_Type *statement_list) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = SCRIPT;
    o->u.statement_list = statement_list;
    return o;
}

Toy_Type*
new_statement(Toy_Type *item_list, int line) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = STATEMENT;
    o->u.statement.item_list = item_list;
    o->u.statement.line = line;
    return o;
}

Toy_Type*
new_block(Toy_Type *block_body) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = BLOCK;
    o->u.block_body = block_body;
    return o;
}

Toy_Type*
new_eval(Toy_Type *eval_body) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = EVAL;
    o->u.eval_body = eval_body;
    return o;
}

Toy_Type*
new_native(Toy_Type* (*cfunc)(Toy_Interp*, Toy_Type*, struct _hash*, int arglen),
	   Toy_Type *argspec_list) {

    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = NATIVE;
    o->u.native.cfunc = cfunc;
    o->u.native.argspec_list = argspec_list;
    return o;
}

Toy_Type*
new_object(char *name, struct _hash* slots, Toy_Type *delegate_list) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = OBJECT;
    o->u.object.slots = slots;
    o->u.object.delegate_list = delegate_list;
    hash_set_t(slots, const_atname, new_symbol(name));
#ifdef HAS_MCACHE
    o->u.object.mcache = new_hash();
#endif
    return o;
}

Toy_Type*
new_exception(char *code, char *desc, Toy_Interp *interp) {
    Toy_Type *o;
    Toy_Type *list;

    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = EXCEPTION;
    o->u.exception.code = new_cell(code);
    o->u.exception.msg_list = new_list(new_string_str(desc));

    if (interp) {
	Cell *stack;

	stack = new_cell(to_string(o));
	cell_add_str(stack, "\n");
	get_stack_trace(interp, stack);

	list_append(o->u.exception.msg_list, new_string_cell(stack));

    } else {

	list_append(o->u.exception.msg_list, new_string_str(""));
    }

    return o;
}

Toy_Type*
new_closure(Toy_Type *block_body, Toy_Env* env, int script_id) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = CLOSURE;
    SET_SCRIPT_ID(block_body, script_id);
    o->u.closure.block_body = block_body;
    o->u.closure.env = env;
    return o;
}

Toy_Type*
new_func(Toy_Type *argspec_list, int posarglen, Array *posargarray, Hash *namedarg, Toy_Type *closure) {
    struct _toy_argspec *argspec;

    argspec = GC_MALLOC(sizeof(Toy_Argspec));
    if (NULL == argspec) return NULL;

    argspec->list = argspec_list;
    argspec->posarg_len = posarglen;
    argspec->posarg_array = posargarray;
    argspec->namedarg = namedarg;

    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = FUNC;
    o->u.func.argspec = argspec;
    o->u.func.closure = closure;
    return o;
}

Toy_Type*
new_control(int code, Toy_Type *ret_value) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = CONTROL;
    o->u.control.code = code;
    o->u.control.ret_value = ret_value;
    return o;
}

Toy_Type*
new_container(void *container) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = CONTAINER;
    o->u.container = container;
    return o;
}

Toy_Type*
new_getmacro(Toy_Type *obj, Toy_Type *para) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = GETMACRO;
    o->u.getmacro.obj = obj;
    o->u.getmacro.para = para;
    return o;
}

Toy_Type*
new_initmacro(Toy_Type *class, Toy_Type *param) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = INITMACRO;
    o->u.initmacro.class = class;
    o->u.initmacro.param = param;
    return o;
}

Toy_Type*
new_alias(struct _hash *slot, Toy_Type *key) {
    Toy_Type *o;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;

    o->tag = ALIAS;
    o->u.alias.slot = slot;
    o->u.alias.key = key;
    return o;
}

Toy_Type*
new_rquote(char *string) {
    Toy_Type *o;
    Cell *c;
    o = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == o) return NULL;
    c = new_cell(string);
    if (NULL == c) return NULL;

    o->tag = RQUOTE;
    o->u.string = c;
    return o;
}

Toy_Type*
toy_clone(Toy_Type *obj) {
    Toy_Type *dest;

    dest = GC_MALLOC(sizeof(Toy_Type));
    if (NULL == dest) return NULL;
    memcpy((void*)dest, (const void*)obj, sizeof(Toy_Type));
    return dest;
}

static char *toy_type_char[] = {
    "NIL",
    "SYMBOL",
    "REF",
    "LIST",
    "INTEGER",
    "REAL",
    "STRING",
    "SCRIPT",
    "STATEMENT",
    "BLOCK",
    "EVAL",
    "NATIVE",
    "OBJECT",
    "EXCEPTION",
    "CLOSURE",
    "FUNC",
    "CONTROL",
    "CONTAINER",
    "GETMACRO",
    "ALIAS",
    "RQUOTE",
    "INITMACRO",
};

char*
toy_get_type_str(Toy_Type *obj) {
    if (NULL == obj) return "(null address)";
    if (IS_TOY_OBJECT(obj)) {
	return toy_type_char[GET_TAG(obj)];
    } else {
	return "(unknown)";
    }
}

char*
to_string(Toy_Type *obj) {
    if (NULL == obj) return "(null address)";
    if (! IS_TOY_OBJECT(obj)) return "(unknown)";

    switch (GET_TAG(obj)) {
    case NIL:
	return S_NIL;

    case SYMBOL:
	return cell_get_addr(obj->u.symbol.cell);

    case REF:
    {
	Cell *c;
	c = new_cell("$");
	cell_add_str(c, cell_get_addr(obj->u.ref.cell));
	return cell_get_addr(c);
    }

    case INTEGER:
    {
	char buff[30];
	Cell *c;
	snprintf(buff, 30, "%lld", obj->u.integer);
	c = new_cell(buff);
	return cell_get_addr(c);
    }

    case REAL:
    {
	char buff[30];
	Cell *c;

	if ((obj->u.real > 1.0E+9) || (obj->u.real < 1.0E-2)) {
	    snprintf(buff, 30, "%.15E", obj->u.real);
	} else {
	    snprintf(buff, 30, "%.15f", obj->u.real);
	}
	c = new_cell(buff);
	return cell_get_addr(c);
    }

    case STRING:
    {
	Cell *c;

	/*
	c = new_cell("\"");
	cell_add_str(c, cell_get_addr(obj->u.string));
	cell_add_char(c, '"');
	return cell_get_addr(c);
	*/

	return cell_get_addr(obj->u.string);
    }
	
    case SCRIPT:
    {
	Toy_Type *l;
	Cell *c;

	l = obj->u.statement_list;
	if (IS_LIST_NULL(l)) {
	    return "";
	}

	c = new_cell("");
	do {
	    cell_add_str(c, to_print(list_get_item(l)));
	} while (l = list_next(l));
	return cell_get_addr(c);
    }

    case STATEMENT:
    {
	Toy_Type *l;
	Cell *c;

	l = obj->u.statement.item_list;
	if (IS_LIST_NULL(l)) {
	    return "";
	}

	c = new_cell("");
	do {
	    cell_add_str(c, to_print(list_get_item(l)));
	    if (list_length(l) > 1) cell_add_char(c, ' ');
	} while (l = list_next(l));
	cell_add_char(c, ';');
	return cell_get_addr(c);
    }

    case LIST:
    {
	Cell *c;

	if (IS_LIST_NULL(obj)) {
	    return "()";
	}

	c = new_cell("(");
	do {
	    if (GET_TAG(obj) != LIST) {
		cell_add_str(c, ". ");
		cell_add_str(c, to_print(obj));
		break;
	    } else {
		cell_add_str(c, to_print(list_get_item(obj)));
		if (list_length(obj) > 1) cell_add_char(c, ' ');
	    }
	} while (obj = list_next(obj));
	cell_add_str(c, ")");
	return cell_get_addr(c);
    }
	
    case BLOCK:
    {
	Toy_Type *s;
	Cell *c;

	s = obj->u.block_body;
	c = new_cell("{");
	cell_add_str(c, to_print(s));
	cell_add_str(c, "}");
	return cell_get_addr(c);
    }

    case EVAL:
    {
	Toy_Type *s;
	Cell *c;

	s = obj->u.eval_body;
	c = new_cell("[");
	cell_add_str(c, to_print(s));
	cell_add_str(c, "]");
	return cell_get_addr(c);
    }

    case NATIVE:
	return "<NATIVE>";

    case EXCEPTION:
    {
	Cell *c;
	c = new_cell("<");
	cell_add_str(c, cell_get_addr(obj->u.exception.code));
	cell_add_str(c, "># ");
	cell_add_str(c, to_string(list_get_item(obj->u.exception.msg_list)));
	return cell_get_addr(c);
    }

    case FUNC:
    {
	Toy_Type *l;
	Cell *c;

	c = new_cell("(");
	l = obj->u.func.argspec->list;
	while (! IS_LIST_NULL(l)) {
	    if (list_get_item(l) == NIL) break;
	    cell_add_str(c, to_print(list_get_item(l)));
	    if (list_length(l) > 1) cell_add_char(c, ' ');
	    l = list_next(l);
	}
	cell_add_str(c, ") ");
	cell_add_str(c, to_print(obj->u.func.closure));
	return cell_get_addr(c);
    }

    case CONTROL:
    {
	char *p;
	Cell *c;
	switch (obj->u.control.code) {
	case CTRL_RETURN:
	    p = "<CONTROL-return># ";
	    break;
	case CTRL_CONTINUE:
	    p = "<CONTROL-continue># ";
	    break;
	case CTRL_BREAK:
	    p = "<CONTROL-break># ";
	    break;
	case CTRL_REDO:
	    p = "<CONTROL-redo># ";
	    break;
	case CTRL_RETRY:
	    p = "<CONTROL-retry># ";
	    break;
	default:
	    p = "<CONTROL-unknown># ";
	}
	c = new_cell(p);
	cell_add_str(c, to_string(obj->u.control.ret_value));
	return cell_get_addr(c);
    }

    case OBJECT:
    {
	Cell *c;
	Hash *h;
	Toy_Type *name;

	h = obj->u.object.slots;
	c = new_cell("(object)");
	name = hash_get_t(h, const_atname);
	if (name) {
	    cell_add_str(c, to_string(name));
	}
	return cell_get_addr(c);
    }

    case CLOSURE:
    {
	Toy_Type *s;
	Cell *c;

	s = obj->u.closure.block_body;
	c = new_cell("{");
	cell_add_str(c, to_print(s));
	cell_add_str(c, "}");
	return cell_get_addr(c);
    }

    case CONTAINER:
	return "<CONTAINER>";

    case GETMACRO:
    {
	Cell *c;

	c = new_cell("");
	cell_add_str(c, to_string(obj->u.getmacro.obj));
	cell_add_str(c, ",");
	cell_add_str(c, to_string(obj->u.getmacro.para));
	return cell_get_addr(c);
    }

    case RQUOTE:
    {
	Cell *c;

	return cell_get_addr(obj->u.rquote);
    }

    case INITMACRO:
    {
	Cell *c;

	c = new_cell("`");
	cell_add_str(c, to_print(obj->u.initmacro.class));
	if (to_print(obj->u.initmacro.param)) {
	    cell_add_str(c, " ");
	    cell_add_str(c, to_print(obj->u.initmacro.param));
	}
	return cell_get_addr(c);
    }

    default:
	return "(unknown)";
    }
}

char*
to_print(Toy_Type *obj) {
    if (NULL == obj) return "(null address)";
    if (! IS_TOY_OBJECT(obj)) return "(unknown)";

    switch (GET_TAG(obj)) {
    case NIL:
	return S_NIL;

    case SYMBOL:
	return cell_get_addr(obj->u.symbol.cell);

    case REF:
    {
	Cell *c;
	c = new_cell("$");
	cell_add_str(c, cell_get_addr(obj->u.ref.cell));
	return cell_get_addr(c);
    }

    case INTEGER:
    {
	char buff[30];
	Cell *c;
	snprintf(buff, 30, "%lld", obj->u.integer);
	c = new_cell(buff);
	return cell_get_addr(c);
    }

    case REAL:
    {
	char buff[30];
	Cell *c;

	if ((obj->u.real > 1.0E+9) || (obj->u.real < 1.0E-2)) {
	    snprintf(buff, 30, "%.15E", obj->u.real);
	} else {
	    snprintf(buff, 30, "%.15f", obj->u.real);
	}
	c = new_cell(buff);
	return cell_get_addr(c);
    }

    case STRING:
    {
	Cell *c;
	char *p;
	c = new_cell("\"");
	p = cell_get_addr(obj->u.string);
	while (*p) {
	    switch (*p) {
	    case '\t':
		cell_add_str(c, "\\t");
		break;
	    case '\r':
		cell_add_str(c, "\\r");
		break;
	    case '\n':
		cell_add_str(c, "\\n");
		break;
		    
	    default:
		cell_add_char(c, *p);
	    }
	    p++;
	}
	cell_add_char(c, '"');
	return cell_get_addr(c);
    }
	
    case SCRIPT:
    {
	Toy_Type *l;
	Cell *c;

	l = obj->u.statement_list;
	if (IS_LIST_NULL(l)) {
	    return "";
	}

	c = new_cell("");
	do {
	    cell_add_str(c, to_print(list_get_item(l)));
	} while (l = list_next(l));
	return cell_get_addr(c);
    }

    case STATEMENT:
    {
	Toy_Type *l;
	Cell *c;

	l = obj->u.statement.item_list;
	if (IS_LIST_NULL(l)) {
	    return "";
	}

	c = new_cell("");
	do {
	    cell_add_str(c, to_print(list_get_item(l)));
	    if (list_length(l) > 1) cell_add_char(c, ' ');
	} while (l = list_next(l));
	cell_add_char(c, ';');
	return cell_get_addr(c);
    }

    case LIST:
    {
	Cell *c;

	if (IS_LIST_NULL(obj)) {
	    return "()";
	}

	c = new_cell("(");
	do {
	    if (GET_TAG(obj) != LIST) {
		cell_add_str(c, ". ");
		cell_add_str(c, to_print(obj));
		break;
	    } else {
		cell_add_str(c, to_print(list_get_item(obj)));
		if (list_length(obj) > 1) cell_add_char(c, ' ');
	    }
	} while (obj = list_next(obj));
	cell_add_str(c, ")");
	return cell_get_addr(c);
    }
	
    case BLOCK:
    {
	Toy_Type *s;
	Cell *c;

	s = obj->u.block_body;
	c = new_cell("{");
	cell_add_str(c, to_print(s));
	cell_add_str(c, "}");
	return cell_get_addr(c);
    }

    case EVAL:
    {
	Toy_Type *s;
	Cell *c;

	s = obj->u.eval_body;
	c = new_cell("[");
	cell_add_str(c, to_print(s));
	cell_add_str(c, "]");
	return cell_get_addr(c);
    }

    case NATIVE:
	return "<NATIVE>";

    case EXCEPTION:
    {
	Cell *c;
	c = new_cell("<");
	cell_add_str(c, cell_get_addr(obj->u.exception.code));
	cell_add_str(c, "># ");
	cell_add_str(c, to_print(list_get_item(obj->u.exception.msg_list)));
	return cell_get_addr(c);
    }

    case FUNC:
    {
	Toy_Type *l;
	Cell *c;

	c = new_cell("(");
	l = obj->u.func.argspec->list;
	while (! IS_LIST_NULL(l)) {
	    if (list_get_item(l) == NIL) break;
	    cell_add_str(c, to_print(list_get_item(l)));
	    if (list_length(l) > 1) cell_add_char(c, ' ');
	    l = list_next(l);
	}
	cell_add_str(c, ") ");
	cell_add_str(c, to_print(obj->u.func.closure));
	return cell_get_addr(c);
    }

    case CONTROL:
    {
	char *p;
	Cell *c;
	switch (obj->u.control.code) {
	case CTRL_RETURN:
	    p = "<CONTROL-return># ";
	    break;
	case CTRL_CONTINUE:
	    p = "<CONTROL-continue># ";
	    break;
	case CTRL_BREAK:
	    p = "<CONTROL-break># ";
	    break;
	case CTRL_REDO:
	    p = "<CONTROL-redo># ";
	    break;
	case CTRL_RETRY:
	    p = "<CONTROL-retry># ";
	    break;
	default:
	    p = "<CONTROL-unknown># ";
	}
	c = new_cell(p);
	cell_add_str(c, to_print(obj->u.control.ret_value));
	return cell_get_addr(c);
    }

    case OBJECT:
    {
	Cell *c;
	Hash *h;
	Toy_Type *name;

	h = obj->u.object.slots;
	c = new_cell("(object)");
	name = hash_get_t(h, const_atname);
	if (name) {
	    cell_add_str(c, to_print(name));
	}
	return cell_get_addr(c);
    }

    case CLOSURE:
    {
	Toy_Type *s;
	Cell *c;

	s = obj->u.closure.block_body;
	c = new_cell("{");
	cell_add_str(c, to_print(s));
	cell_add_str(c, "}");
	return cell_get_addr(c);
    }

    case CONTAINER:
	return "<CONTAINER>";

    case GETMACRO:
    {
	Cell *c;

	c = new_cell("");
	cell_add_str(c, to_print(obj->u.getmacro.obj));
	cell_add_str(c, ",");
	cell_add_str(c, to_print(obj->u.getmacro.para));
	return cell_get_addr(c);
    }

    case RQUOTE:
    {
	Cell *c;
	char *p;
	c = new_cell("\'");
	p = cell_get_addr(obj->u.rquote);
	while (*p) {
	    switch (*p) {
	    case '\\':
		cell_add_str(c, "\\\\");
		break;
	    case '\'':
		cell_add_str(c, "\\\'");
		break;
	    default:
		cell_add_char(c, *p);
	    }
	    p++;
	}
	cell_add_char(c, '\'');
	return cell_get_addr(c);
    }

    case INITMACRO:
    {
	Cell *c;

	c = new_cell("`");
	cell_add_str(c, to_print(obj->u.initmacro.class));
	if (obj->u.initmacro.param) {
	    cell_add_str(c, " ");
	    cell_add_str(c, to_print(obj->u.initmacro.param));
	}
	return cell_get_addr(c);
    }

    default:
	return "(unknown)";
    }
}
